@@ -2,13 +2,12 @@
 //!
 use std::collections::{HashMap, HashSet};
 use std::hash::{Hash, Hasher};
-use std::marker::PhantomData;
 
 use ndarray::{Array1, ArrayBase, Axis, Data, Ix1, Ix2};
 
-use super::hyperparameters::{DecisionTreeParams, SplitQuality};
 use super::NodeIter;
 use super::Tikz;
+use super::{DecisionTreeValidParams, SplitQuality};
 use linfa::{
     dataset::{AsTargets, Labels, Records},
     error::Error,
@@ -200,7 +199,7 @@ impl<F: Float, L: Label + std::fmt::Debug> TreeNode<F, L> {
     fn fit<D: Data<Elem = F>, T: AsTargets<Elem = L> + Labels<Elem = L>>(
         data: &DatasetBase<ArrayBase<D, Ix2>, T>,
         mask: &RowMask,
-        hyperparameters: &DecisionTreeParams<F, L>,
+        hyperparameters: &DecisionTreeValidParams<F, L>,
         sorted_indices: &[SortedIndex<F>],
         depth: usize,
     ) -> Result<Self> {
@@ -212,9 +211,9 @@ impl<F: Float, L: Label + std::fmt::Debug> TreeNode<F, L> {
         let target = data.try_single_target()?;
 
         // return empty leaf when we don't have enough samples or the maximal depth is reached
-        if (mask.nsamples as f32) < hyperparameters.min_weight_split
+        if (mask.nsamples as f32) < hyperparameters.min_weight_split()
             || hyperparameters
-                .max_depth
+                .max_depth()
                 .map(|max_depth| depth >= max_depth)
                 .unwrap_or(false)
         {
@@ -277,14 +276,14 @@ impl<F: Float, L: Label + std::fmt::Debug> TreeNode<F, L> {
 
                 // If the split would result in too few samples in a leaf
                 // then skip computing the quality
-                if weight_on_right_side < hyperparameters.min_weight_leaf
-                    || weight_on_left_side < hyperparameters.min_weight_leaf
+                if weight_on_right_side < hyperparameters.min_weight_leaf()
+                    || weight_on_left_side < hyperparameters.min_weight_leaf()
                 {
                     continue;
                 }
 
                 // Calculate the quality of each resulting subset of the dataset
-                let (left_score, right_score) = match hyperparameters.split_quality {
+                let (left_score, right_score) = match hyperparameters.split_quality() {
                     SplitQuality::Gini => (
                         gini_impurity(&right_class_freq),
                         gini_impurity(&left_class_freq),
@@ -321,7 +320,7 @@ impl<F: Float, L: Label + std::fmt::Debug> TreeNode<F, L> {
         // a leaf node that predicts the most common label in the available observations.
 
         let impurity_decrease = if let Some((_, _, best_score)) = best {
-            let parent_score = match hyperparameters.split_quality {
+            let parent_score = match hyperparameters.split_quality() {
                 SplitQuality::Gini => gini_impurity(&parent_class_freq),
                 SplitQuality::Entropy => entropy(&parent_class_freq),
             };
@@ -334,7 +333,7 @@ impl<F: Float, L: Label + std::fmt::Debug> TreeNode<F, L> {
             F::zero()
         };
 
-        if impurity_decrease < hyperparameters.min_impurity_decrease {
+        if impurity_decrease < hyperparameters.min_impurity_decrease() {
             return Ok(Self::empty_leaf(prediction, depth));
         }
 
@@ -359,8 +358,8 @@ impl<F: Float, L: Label + std::fmt::Debug> TreeNode<F, L> {
             Some(Box::new(TreeNode::fit(
                 data,
                 &left_mask,
-                &hyperparameters,
-                &sorted_indices,
+                hyperparameters,
+                sorted_indices,
                 depth + 1,
             )?))
         } else {
@@ -371,8 +370,8 @@ impl<F: Float, L: Label + std::fmt::Debug> TreeNode<F, L> {
             Some(Box::new(TreeNode::fit(
                 data,
                 &right_mask,
-                &hyperparameters,
-                &sorted_indices,
+                hyperparameters,
+                sorted_indices,
                 depth + 1,
             )?))
         } else {
@@ -489,20 +488,29 @@ pub struct DecisionTree<F: Float, L: Label> {
     num_features: usize,
 }
 
-impl<F: Float, L: Label, D: Data<Elem = F>> PredictRef<ArrayBase<D, Ix2>, Array1<L>>
+impl<F: Float, L: Label + Default, D: Data<Elem = F>> PredictInplace<ArrayBase<D, Ix2>, Array1<L>>
     for DecisionTree<F, L>
 {
     /// Make predictions for each row of a matrix of features `x`.
-    fn predict_ref(&self, x: &ArrayBase<D, Ix2>) -> Array1<L> {
-        x.genrows()
-            .into_iter()
-            .map(|row| make_prediction(&row, &self.root_node))
-            .collect()
+    fn predict_inplace(&self, x: &ArrayBase<D, Ix2>, y: &mut Array1<L>) {
+        assert_eq!(
+            x.nrows(),
+            y.len(),
+            "The number of data points must match the number of output targets."
+        );
+
+        for (row, target) in x.rows().into_iter().zip(y.iter_mut()) {
+            *target = make_prediction(&row, &self.root_node);
+        }
+    }
+
+    fn default_target(&self, x: &ArrayBase<D, Ix2>) -> Array1<L> {
+        Array1::default(x.nrows())
     }
 }
 
 impl<'a, F: Float, L: Label + 'a + std::fmt::Debug, D, T> Fit<ArrayBase<D, Ix2>, T, Error>
-    for DecisionTreeParams<F, L>
+    for DecisionTreeValidParams<F, L>
 where
     D: Data<Elem = F>,
     T: AsTargets<Elem = L> + Labels<Elem = L>,
@@ -512,18 +520,16 @@ where
     /// Fit a decision tree using `hyperparamters` on the dataset consisting of
     /// a matrix of features `x` and an array of labels `y`.
     fn fit(&self, dataset: &DatasetBase<ArrayBase<D, Ix2>, T>) -> Result<Self::Object> {
-        self.validate()?;
-
         let x = dataset.records();
         let feature_names = dataset.feature_names();
         let all_idxs = RowMask::all(x.nrows());
         let sorted_indices: Vec<_> = (0..(x.ncols()))
             .map(|feature_idx| {
-                SortedIndex::of_array_column(&x, feature_idx, &feature_names[feature_idx])
+                SortedIndex::of_array_column(x, feature_idx, &feature_names[feature_idx])
             })
             .collect();
 
-        let mut root_node = TreeNode::fit(&dataset, &all_idxs, &self, &sorted_indices, 0)?;
+        let mut root_node = TreeNode::fit(dataset, &all_idxs, self, &sorted_indices, 0)?;
         root_node.prune();
 
         Ok(DecisionTree {
@@ -533,26 +539,7 @@ where
     }
 }
 
-impl<F: Float, L: Label + std::fmt::Debug> DecisionTree<F, L> {
-    /// Defaults are provided if the optional parameters are not specified:
-    /// * `split_quality = SplitQuality::Gini`
-    /// * `max_depth = None`
-    /// * `min_weight_split = 2.0`
-    /// * `min_weight_leaf = 1.0`
-    /// * `min_impurity_decrease = 0.00001`
-    // Violates the convention that new should return a value of type `Self`
-    #[allow(clippy::new_ret_no_self)]
-    pub fn params() -> DecisionTreeParams<F, L> {
-        DecisionTreeParams {
-            split_quality: SplitQuality::Gini,
-            max_depth: None,
-            min_weight_split: 2.0,
-            min_weight_leaf: 1.0,
-            min_impurity_decrease: F::cast(0.00001),
-            phantom: PhantomData,
-        }
-    }
-
+impl<F: Float, L: Label> DecisionTree<F, L> {
     /// Create a node iterator in level-order (BFT)
     pub fn iter_nodes(&self) -> NodeIter<F, L> {
         // queue of nodes yet to explore
@@ -633,7 +620,7 @@ impl<F: Float, L: Label + std::fmt::Debug> DecisionTree<F, L> {
     /// * `complete=true`
     ///
     pub fn export_to_tikz(&self) -> Tikz<F, L> {
-        Tikz::new(&self)
+        Tikz::new(self)
     }
 }
 
@@ -706,7 +693,7 @@ mod tests {
     use super::*;
 
     use approx::assert_abs_diff_eq;
-    use linfa::{error::Result, metrics::ToConfusionMatrix, Dataset};
+    use linfa::{error::Result, metrics::ToConfusionMatrix, Dataset, ParamGuard};
     use ndarray::{array, concatenate, s, Array, Array1, Array2, Axis};
     use rand::rngs::SmallRng;
 
@@ -916,7 +903,7 @@ mod tests {
     fn panic_min_impurity_decrease() {
         DecisionTree::<f64, bool>::params()
             .min_impurity_decrease(0.0)
-            .validate()
+            .check()
             .unwrap();
     }
 }
